#!/usr/bin/python3

import time
import re
import os
import shlex
import requests
import datetime
import zipfile
import shutil
import urllib.parse
import subprocess
from selenium import webdriver
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.common.by import By
from bs4 import BeautifulSoup

from addons import ADDONS, ZIP_DIRECTORY, ADDON_DIRECTORY, ZIP_FILE_NAME_PATTERN, PATCH_DIRECTORY, get_current_version

USER_AGENT = "Mozilla/5.0 (X11; Linux x86_64; rv:105.0) Gecko/20100101 Firefox/105.0"
BASE_URL = "https://www.curseforge.com"


def retrieve_last_update_ts(soup):

    div = soup.find("div", {"class": "cf-sidebar-inner"})
    if not div:
        return None
    
    elements = div.find_all(recursive=False)
    for i, element in enumerate(elements):
        if element.name == "h4" and element.text.strip() == "WoW Wrath of the Lich King Classic":
            if i + 1 < len(elements) and elements[i+1].name == "ul":
                try:
                    abbr = elements[i+1].find("abbr")
                    a = elements[i+1].find("a", {"data-tooltip": "Download file"})
                    epoch = abbr["data-epoch"]
                    if epoch:
                        return datetime.datetime.fromtimestamp(int(epoch)), a["href"]
                except Exception as ex:
                    print(ex)
                    pass
    
    print("Element not found")


def wait_for_download(driver, name):
    driver.execute_script("window.open('');")
    time.sleep(1)
    driver.switch_to.window(driver.window_handles[-1])
    driver.get("chrome://downloads/")
    downloaded_file = None
    while not downloaded_file:
        files = driver.execute_script("""
            var downloadsManager = document.querySelector('downloads-manager');
            if (downloadsManager) {
                var items = downloadsManager.shadowRoot.getElementById('downloadsList').items;
                if (items.every(e => e.state === "COMPLETE"))
                    return items.map(e => e.fileUrl || e.file_url);
            }""")
        
        if files:
            for file in files:
                parts = urllib.parse.urlparse(file)
                if parts.scheme == "file":
                    abs_path = urllib.parse.unquote(parts.path)
                    if os.path.isfile(abs_path):
                        if re.findall(name + ZIP_FILE_NAME_PATTERN, os.path.basename(abs_path), flags=re.IGNORECASE):
                            downloaded_file = abs_path
                            break

        time.sleep(1)
    driver.execute_script("window.close();")
    driver.switch_to.window(driver.window_handles[0])
    return downloaded_file

def extract_update(path):
    zip_file = zipfile.ZipFile(path, "r")
    directories = set()

    for entry in zip_file.namelist():
        if "/" in entry and not entry.startswith("/") and not ".." in entry:
            directories.add(entry.split("/")[0])

    for directory in directories:
        if os.path.isdir(directory):
            shutil.rmtree(directory)
    
    zip_file.extractall(ADDON_DIRECTORY)

    for directory in directories:
        patch_file = os.path.join(PATCH_DIRECTORY, f"{directory}.patch")
        if os.path.isfile(patch_file):
            print("Applying patch:", patch_file)
            patch_file_arg = shlex.quote(os.path.realpath(patch_file))
            cwd = os.path.realpath(directory)
            proc = subprocess.Popen(["sh", "-c", f"patch -p0 < {patch_file_arg}"], cwd=cwd)
            proc.wait()

    zip_file.close()

def update_addon(driver, url, name):

    current_version = get_current_version(name)
    driver.get(f"{BASE_URL}/wow/addons/{url}")
    WebDriverWait(driver, 3).until(EC.presence_of_element_located((By.TAG_NAME, 'body')))

    soup = BeautifulSoup(driver.page_source, "html.parser")
    check = retrieve_last_update_ts(soup)
    if not check:
        print(f"[-] Unable to get last updated timestamp for: {addon}")
        return False
    
    last_updated, download_url = check
    ts = last_updated.strftime("%d.%m.%Y %H:%M")
    current = "N/A" if current_version is None else current_version.strftime("%d.%m.%Y %H:%M")
    print(f"[+] {addon} last updated @ {ts}, current @ {current}")
    if current_version is None or last_updated > current_version:
        print("[~] Downloading new file…")
        driver.get(BASE_URL + download_url)
        path = wait_for_download(driver, addon)
        print("[+] Download completed:", os.path.basename(path))
        extract_update(path)
        print(f"[+] {addon} Updated successfully!")

if __name__ == "__main__":

    if not os.path.isdir(ADDON_DIRECTORY):
        print("AddOns directory not found. Are you executing this script from the correct path?")
    else:

        options = Options()
        # options.headless = True
        options.add_argument(f"user-agent={USER_AGENT}")
        prefs = {
            "download.default_directory": os.path.realpath(ZIP_DIRECTORY),
            "profile.managed_default_content_settings.images": 2
        }
        options.add_experimental_option("prefs", prefs)

        driver = webdriver.Chrome(options=options)
        for url, addon in ADDONS.items():
            update_addon(driver, url, addon)
        driver.close()
